  #include <Arduino.h>
  #include "SamcoPositionSimplified.h"

  #define MANUFACTURER_NAME "3DP"
  #define DEVICE_NAME       "GUN4ALL-Con"
  #define DEVICE_VID        0x0920
  #define DEVICE_PID        0x1998  
  #define DFROBOT_IR_CLOCK  1000000 

  namespace IRCAM {
    bool Poll();
    void Trigger(int pressed);
    void GetPosition();
    void TinyUSBInit();
    void startIrCamTimer(int frequencyHz);
    void rp2040pwmIrq(void);
    void rp2040EnablePWMTimer(unsigned int slice_num, unsigned int frequency);
    /* ------------------------------------------------------------------------- */
    
    DFRobotIRPositionEx::Sensitivity_e irSensitivity = DFRobotIRPositionEx::Sensitivity_Default;
    
    constexpr unsigned int IRCamUpdateRate = 209; 
    constexpr float ScaleStep = 0.001; 
    unsigned long IRCAMpollTimer;
    volatile unsigned int irPosUpdateTick = 0;
  
    int xCenter = MouseMaxX / 2;
    int yCenter = MouseMaxY / 2;
    float xScale = 1.64;
    float yScale = 0.95;
    
    // Values after tilt correction
    int finalX = 0;         
    int finalY = 0;
    
    // Unconstrained mouse postion
    int moveXAxis = 0;      
    int moveYAxis = 0;               
    int moveXAxisArr[3] = {0, 0, 0};
    int moveYAxisArr[3] = {0, 0, 0};
    int moveIndex = 0;
    
    // Constrained mouse postion
    int conMoveXAxis = 0;   
    int conMoveYAxis = 0;
    
    bool offXAxis = false; 
    bool offYAxis = false;
    
    unsigned int lastSeen = 0;
    
    /* - - - - - - - - - - - - - - - */
  
    void Init() {
      TinyUSBInit();
      
      Wire.setSDA(PIN_SDA);
      Wire.setSCL(PIN_SCL);
      
      DFRobotIRCam  .begin(DFROBOT_IR_CLOCK, DFRobotIRPositionEx::DataFormat_Basic, irSensitivity);  
      TinyUSBDevices.begin(2);
      MOUSE::Pointer.init(MouseMaxX, MouseMaxY, true);
      
      while(!USBDevice.mounted()) { yield(); }
      startIrCamTimer(IRCamUpdateRate);
      
      Keyboard.releaseAll();
      delay(1);
      
      MOUSE::Pointer.release(MOUSE_LEFT);
      MOUSE::Pointer.release(MOUSE_RIGHT);
      MOUSE::Pointer.release(MOUSE_MIDDLE);
      
      IRCAMpollTimer = millis();
      moveIndex = 0; lastSeen = 0x0F;
    }
  
    bool Poll() {
      if(irPosUpdateTick) {
        irPosUpdateTick = 0;
        GetPosition();
        
        int halfHscale = (int)(SamcoPosition.h() * xScale + 0.5f) / 2;
        moveXAxis = map(finalX, xCenter + halfHscale, xCenter - halfHscale, 0, MouseMaxX);
        halfHscale = (int)(SamcoPosition.h() * yScale + 0.5f) / 2;
        moveYAxis = map(finalY, yCenter + halfHscale, yCenter - halfHscale, 0, MouseMaxY);
        
        conMoveXAxis = constrain(moveXAxis, 0, MouseMaxX);
        conMoveYAxis = constrain(moveYAxis, 0, MouseMaxY);  
        
        if (conMoveXAxis == 0 || conMoveXAxis == MouseMaxX) offXAxis = true; else offXAxis = false;
        if (conMoveYAxis == 0 || conMoveYAxis == MouseMaxY) offYAxis = true; else offYAxis = false;
        
        MOUSE::Pointer.move(conMoveXAxis, conMoveYAxis);
       }  
       return (!offXAxis && !offYAxis);
    }
  
    void GetPosition() {
      int error = DFRobotIRCam.basicAtomic(DFRobotIRPositionEx::Retry_2);
      
      if(error == DFRobotIRPositionEx::Error_Success) {
        SamcoPosition.begin(DFRobotIRCam.xPositions(), DFRobotIRCam.yPositions(), DFRobotIRCam.seen(), xCenter, yCenter);
        
        finalX = SamcoPosition.x();
        finalY = SamcoPosition.y();
  
        if (lastSeen != SamcoPosition.seen()) lastSeen = SamcoPosition.seen();

       /*
        #if DEBUG
          if ((!offXAxis && !offYAxis) && (millis()-IRCAMpollTimer)>1000) {
            Serial.print(finalX); Serial.print(F(' '));
            Serial.print(finalY); Serial.print(F("   "));
            Serial.println(SamcoPosition.h());Serial.flush();   
            IRCAMpollTimer = millis(); //update(IRCAMpollTimer);        
          }
        #endif */
        
      } else if(error != DFRobotIRPositionEx::Error_DataMismatch) {
          Serial.println(F("Device not available!"));
      }
    }
  
    void Trigger(int pressed = -1) {
      #ifndef DISABLE_CLICK
        if (pressed != -1) {
          if (pressed == 1) MOUSE::Pointer.press(MOUSE_LEFT);
          if (pressed == 0) MOUSE::Pointer.release(MOUSE_LEFT);
        }
      #endif
    }
  
    void TinyUSBInit() {
      TinyUSBDevice.setManufacturerDescriptor(MANUFACTURER_NAME);
      TinyUSBDevice.setProductDescriptor(DEVICE_NAME);
      TinyUSBDevice.setID(DEVICE_VID, DEVICE_PID);
    } 
  
    void startIrCamTimer(int frequencyHz) {
      rp2040EnablePWMTimer(0, frequencyHz);
      irq_set_exclusive_handler(PWM_IRQ_WRAP, rp2040pwmIrq);
      irq_set_enabled(PWM_IRQ_WRAP, true);
    }
    
    void rp2040EnablePWMTimer(unsigned int slice_num, unsigned int frequency) {
      pwm_config pwmcfg = pwm_get_default_config();
      float clkdiv = (float)clock_get_hz(clk_sys) / (float)(65535 * frequency);
      
      if (clkdiv < 1.0f) clkdiv = 1.0f; 
      else clkdiv += 2.0f / (float)(1u << PWM_CH1_DIV_INT_LSB);
      
      pwm_config_set_clkdiv(&pwmcfg, clkdiv);
      clkdiv = (float)pwmcfg.div / (float)(1u << PWM_CH1_DIV_INT_LSB);
      pwm_config_set_wrap(&pwmcfg, (float)clock_get_hz(clk_sys) / (frequency * clkdiv));
      pwm_init(slice_num, &pwmcfg, true); pwm_set_irq_enabled(slice_num, true);
    }
    
    void rp2040pwmIrq(void) {
      pwm_hw->intr = 0xff;
      irPosUpdateTick = 1;
    }
  }
