#include <Arduino.h>
#include <LiquidCrystal.h>
#include "LCD.h"

#define BL_LIMIT   30000
#define CURSOR_T   500
#define SCROLL_T   250
#define AIN_POLL   50
#define AIN_JITT   20

#define BACKLIGHT  6
#define LED_RED    9
#define LED_GREEN  10

#define POT1       A0
#define POT2       A1
#define SW         A2

#define BTN_LEFT   A3 
#define BTN_RIGHT  A4
#define BTN_MID    A5

unsigned long 
led_ms,
delay_ms,
bl_timer,
bl_update;

String
storedMsg[2],

//String* 
*rows[2] = { 
  &storedMsg[0], 
  &storedMsg[1] 
};

enum : int { POLL, MEAN, PREV };

uint16_t 
analogVals[][3] = {
  {0, 0, 0}, 
  {0, 0, 0}, 
  {0, 0, 0}
};

bool 
bl_timeout = false,
bl_state   = true,
cur_state  = false,
printed    = false;

byte
analogCount = 0,
btnPressed  = 0,
BL_VAL      = 255;

void initInterrupt();
void readSerial();
void readAnalog();
void printMsg();
void checkBL();
void splash3();
void splash2();
void splash();
bool checkBtn();

void toggleBL(bool state);
void cursorToggle(bool scroll);
void cursorBlink(byte blinks, bool scrolling);
bool delay_func(uint16_t del, bool abort);

// ________________________________________________________________

void setup() {
  delay(250); // wait for Serial port to open
  Serial.begin(9600);

  pinMode(SW,   INPUT_PULLUP);
  pinMode(POT1, INPUT_PULLUP);
  pinMode(POT2, INPUT_PULLUP);

  pinMode(BTN_LEFT,    INPUT); digitalWrite(BTN_LEFT,  HIGH);
  pinMode(BTN_RIGHT,   INPUT); digitalWrite(BTN_RIGHT, HIGH);
  pinMode(BTN_MID,     INPUT); digitalWrite(BTN_MID,   HIGH);
  
  pinMode(LED_RED,    OUTPUT); digitalWrite(LED_RED,   HIGH);
  pinMode(LED_GREEN,  OUTPUT); digitalWrite(LED_GREEN, HIGH);
  pinMode(BACKLIGHT,  OUTPUT); digitalWrite(BACKLIGHT, HIGH);
  
  initInterrupt();
 
  lcd_init();
  lcd.clear();

  storedMsg[0] = F("ARCADE LCD:");
  storedMsg[1] = F("SERIAL READY.");
  
  toggleBL(1);
  splash();
  Serial.println(F("LCD ONLINE;")); //Serial.flush();
}

void loop() {
  while (!Serial.available()) 
    printMsg();
    
  if (Serial.available()) 
    readSerial();
}

void readSerial() {
  String input;
  
  while (Serial.available()) input = Serial.readStringUntil(';');

  if (input.length() > 2) { 
    if (isDigit(input[0])) { 
      int row = input.substring(0,1).toInt();
      
      digitalWrite(LED_GREEN, 0); 
      
      if (row >= 0 && row < 2) 
        storedMsg[row] = input.substring(2, input.length()); 
        storedMsg[row].trim();
        
      if (storedMsg[1].length() > 16) storedMsg[1] += " "; // ADD SPACE FOR CURSOR
      toggleBL(1);
      printed = false;
    }
  }
}

void printMsg() {
  byte 
  l, p,
  r_len[2] =  { rows[0]->length(), rows[1]->length() },
  maxP     =  ((r_len[0] >= r_len[1] ) ? r_len[0] : r_len[1]);
        
  if (!printed) { lcd.clear();
    for (l = 0; l < 2; l++) 
      lcd.setCursor(0, l), 
      lcd.print(rows[l]->substring(0,16));
    printed = true;  
  }
  
  cursorBlink(1,0);
  digitalWrite(LED_GREEN, 1);
  
  if (maxP > 16) { 
    for (p = 0; p <= (maxP-16); p++) {
      for (l = 0; l < 2; l++) {
        if ((p + 16) <= r_len[l]) 
          lcd.setCursor(0, l), 
          lcd.print(rows[l]->substring(p, r_len[l]));
      }
      if (!delay_func(SCROLL_T,1)) return;
    }
    cursorBlink(2,(r_len[1] > 16));
    printed = false; // show initial rows with delay if scrolling
  } 
}

bool delay_func(uint16_t del, bool abort = false) {
  bool rc = true;
  
  delay_ms = millis();
  
  while ((millis() - delay_ms) < del) {
    if (Serial.available() && abort) { rc = false; break; }
    if (checkBtn()) delay_ms = millis();  
    if (!digitalRead(LED_RED) && (millis()-led_ms) > 1000) 
      digitalWrite(LED_RED, 1), led_ms = millis();
    
    checkBL(); 
    readAnalog();
  }
  return rc;
}

void readAnalog() {
  int  pins[3]    = {SW, POT1, POT2};
  bool pollAnalog = true;
  
  if (analogCount >= AIN_POLL) pollAnalog = false;
  else 
    for (byte x=0;x<3;x++) 
      if (analogVals[POLL][x] >= 65535) pollAnalog = false;
  
  if (pollAnalog) { analogCount++;
    for (byte x = 0; x < 3; x++) analogVals[POLL][x] += analogRead(pins[x]); 
  } 
  else {
    for (byte x = 0; x < 3; x++) {
      analogVals[MEAN][x] = (analogVals[POLL][x]/analogCount);
      
      if (getDiff(analogVals[MEAN][x], analogVals[PREV][x]) > AIN_JITT) 
        digitalWrite(LED_RED, 0), led_ms = millis(),
        
        Serial.print("AIN "), Serial.print(x), 
        Serial.print(" "),    Serial.print(analogVals[MEAN][x]), 
        Serial.println(";");  //Serial.flush();
        
      analogVals[PREV][x] = analogVals[MEAN][x]; 
      analogVals[POLL][x] = 0;
    }
    analogCount = 0;
  }
}

bool checkBtn() {
  bool pressed = 0;
  cli();
  if (btnPressed > 0) {
    digitalWrite(LED_RED, 0); led_ms = millis();
    Serial.print("BTN "), Serial.print(btnPressed), 
    Serial.println(";"); //Serial.flush();
    btnPressed = 0; delay(100);
  }
  sei();
  return pressed;
}  

void cursorBlink(byte blinks=1, bool scrolling=false) {
  cur_state = 0;
  for (byte n=0; n<(blinks*2); n++) {
    cursorToggle(scrolling);
    if (!delay_func(CURSOR_T,1)) return;    
  }  
}

void cursorToggle(bool scroll=false) {
  bool cursor_print = true;

  if (!scroll && (rows[1]->length() == 16)) 
    cursor_print = false;
    
  if (cursor_print) 
    lcd.setCursor( ((!scroll) ? rows[1]->length() : 15), 1),
    cur_state = !cur_state;
  
  if (cur_state) 
    lcd.write(byte(5)); 
  else 
    lcd.print(" ");
}

void toggleBL(bool state) {
  bl_state = state; 
  BL_VAL = ((state) ? 255 : 0);
  
  analogWrite(BACKLIGHT, BL_VAL); 
  
  if (state) 
    bl_timeout = false, 
    bl_timer =  millis(); 
}

void checkBL() {
  byte fadeTime;
  
  if (bl_state) {
    if (!bl_timeout) {
      if ((millis()-bl_timer) > BL_LIMIT) 
        bl_timeout = true;
    } 
    else {
      if (BL_VAL > 0)  {
        fadeTime = map(BL_VAL, 255, 0, 10, 250);
        if ((millis()-bl_update) > fadeTime) 
          analogWrite(BACKLIGHT, BL_VAL--), 
          bl_update = millis();
      } 
      else { 
        toggleBL(0); bl_timeout = false;
      } 
    }    
  }
}

void splash() {
  byte col = 0, del = 10;

  // WRITE LINES (TOP ROW)
  for (byte r = 0; r<16; r++) {
    for (byte c = 6; c<9; c++) 
      lcd.setCursor(r,0), lcd.write(byte(c)), 
      delay(del); 
  } 

  // WRITE LINES (BOTTOM ROW REVERSE)
  for (byte r = 16; r>0; r--) {
    for (byte c = 10; c>7; c--) 
      lcd.setCursor((r-1),1), lcd.write(byte(c)), 
      delay(del); 
  } 

  // WRITE BLOCKS (BOTH LINES)
  for (byte r = 0; r<16; r++) {
    lcd.setCursor(r,0), lcd.write(byte(5)); 
    lcd.setCursor(r,1), lcd.write(byte(5));
    delay(del*5); 
  } 
  
  // CLEAR LINES
  for (byte r = 0; r<16; r++) {
    for (byte c = 10; c>7; c--) 
      lcd.setCursor(r,0), lcd.write(byte(c)), 
      lcd.setCursor(r,1), lcd.write(byte(c)), 
      delay(del); 
    lcd.setCursor(r,0),lcd.print(" "); 
    lcd.setCursor(r,1),lcd.print(" ");
  } 

  delay(250);
  toggleBL(1);
}

uint16_t getDiff( uint16_t X, uint16_t Y ) {
  int16_t diff = (X - Y);       // Signed difference
  if (diff < 0) diff = (Y - X); // Positive difference
  return diff;                  // Convert back to unsigned
}

void initInterrupt() {
  cli(); 
  PCICR =0x02;
  PCMSK1 = 0b00111000; // A3 - A5
  sei();
}

ISR(PCINT1_vect) {
  if (digitalRead(BTN_LEFT)  == 0) btnPressed = 1;
  if (digitalRead(BTN_MID)   == 0) btnPressed = 2;
  if (digitalRead(BTN_RIGHT) == 0) btnPressed = 3;
}
